""" mcdutils proxy

$Id: proxy.py,v 1.11 2006/06/07 05:42:57 tseaver Exp $
"""
import memcache
import UserDict

from zope.interface import implementedBy
from zope.interface import implements

from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.mcdutils.interfaces import IMemCacheProxy
from Products.mcdutils.mapping import MemCacheMapping

class FauxClient(UserDict.UserDict):

    def _get_server(self, key):
        return self, key

    def set(self, key, value):
        self[key] = value

class MemCacheProxy(SimpleItem, PropertyManager):
    """ Implement ISDC via a a pool of memcache servers.
    """
    implements(IMemCacheProxy
             + implementedBy(SimpleItem)
             + implementedBy(PropertyManager)
             )
    security = ClassSecurityInfo()

    _v_cached = None
    _v_client = None

    security.declarePrivate('_get_cached')
    def _get_cached(self):
        if self._v_cached is None:
            self._v_cached = {}
        return self._v_cached

    _cached = property(_get_cached,)

    security.declarePrivate('_get_client')
    def _get_client(self):
        if self._v_client is not None:
            return self._v_client

        if self.servers:
            client = self._v_client = memcache.Client(self.servers)
        else:
            client = self._v_client = FauxClient()

        return client

    client = property(_get_client,)

    _servers = ()

    security.declarePrivate('_set_servers')
    def _set_servers(self, value):
        self._servers = value
        try:
            del self._v_client
        except AttributeError:
            pass
        try:
            del self._v_cache
        except AttributeError:
            pass

    servers = property(lambda self: self._servers, _set_servers)

    #
    #   ZMI
    #
    meta_type = 'MemCache Proxy'
    _properties = (
        {'id': 'servers', 'type': 'lines', 'mode': 'w'},
    )

    manage_options = (PropertyManager.manage_options
                    + SimpleItem.manage_options
                     )

    security.declarePrivate('get')
    def get(self, key):
        """ See IMemCacheProxy.
        """
        mapping = self._cached.get(key)

        if mapping is None:
            mapping = self._get_remote(key)

        if mapping is not None:
            # Force this mapping to clean up after the transaction.
            mapping.register(mapping)

        return mapping

    security.declarePrivate('get_multi')
    def get_multi(self, keys):
        """ See IMemCacheProxy.
        """
        return self._get_remote_multi(keys)

    security.declarePrivate('set')
    def set(self, key, value):
        """ See IMemCacheProxy.
        """
        rc = self.client.set(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('add')
    def add(self, key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key does not already exist.
        """
        rc = self.client.add(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('replace')
    def replace(self, key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key already exists.
        """
        rc = self.client.replace(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('delete')
    def delete(self, key, time=0):
        """ Remove the value stored in the cache under 'key'.

        o Return a boolean to indicate success.

        o 'time', if nonzero, will be tested in the cache to determine
          whether the item's 
        """
        rc = self.client.delete(key, time)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('create')
    def create(self, key):
        """ See IMemCacheProxy.
        """
        mapping = self._cached[key] = MemCacheMapping(key, self)
        return mapping

    #
    #   Helper methods
    #
    security.declarePrivate('_get_remote')
    def _get_remote(self, key):
        mapping = self._cached[key] = self.client.get(key)
        if isinstance(mapping, MemCacheMapping):
            mapping._p_proxy = self
            mapping._p_key = key
            mapping._p_oid = hash(key)
            mapping._p_jar = mapping  
            mapping._p_joined = False
            mapping._p_changed = 0
        return mapping

    security.declarePrivate('_get_remote_multi')
    def _get_remote_multi(self, keys):
        meta_map = self.client.get_multi(keys)
        result = {}
        for key, mapping in meta_map.items():
            if isinstance(mapping, MemCacheMapping):
                mapping._p_proxy = self
                mapping._p_key = key
                mapping._p_oid = hash(key)
                mapping._p_jar = mapping  
                mapping._p_joined = False
                mapping._p_changed = 0
            result[k] = mapping

        return result

InitializeClass(MemCacheProxy)


def addMemCacheProxy(dispatcher, id, REQUEST):
    """ Add a MCP to dispatcher.
    """
    proxy = MemCacheProxy()
    proxy._setId(id)
    dispatcher._setObject(id, proxy)
    REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                    % dispatcher.absolute_url())

addMemCacheProxyForm = PageTemplateFile('www/add_mcp.pt', globals())



""" RAMCacheManager workalike using memcache

$Id: zcache.py,v 1.1 2006/06/07 05:43:09 tseaver Exp $
"""
from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass
from OFS.Cache import CacheManager
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.interface import implementedBy
from zope.interface import implements

from Products.mcdutils.interfaces import IZCache
from Products.mcdutils.interfaces import IZCacheManager

def aggregateKey(ob,
                 view_name='',
                 request=None,
                 request_names=(),
                 local_keys=None,
                ):
    """ Return a key to be used when retrieving or inserting a cache entry.

    o 'ob' is the object for whom the key is desired.

    o 'view_name' must be a string.

    o 'request' must be a mapping.

    o 'request_names' must be a sorted list of keys to be included in the key
      from 'request'

    o 'local_keys' is a mapping or None.
    """
    path = '/'.join(ob.getPhysicalPath())
    request_index = []
    local_index = []
    if request is None:
        request = {}
    if local_keys is None:
        local_keys = {}

    for key in request_names:
        val = request.get(key, '')
        request_index.append('%s:%s' % (str(key), str(val)))

    for key, val in local_keys.items():
        local_index.append('%s:%s' % (str(key), str(val)))
    local_index.sort()

    return '|'.join((path,
                     str(view_name),
                     ','.join(request_index),
                     ','.join(local_index),
                    ))

class MemCacheZCache(object):
    """ Implement ISDC via a memcache proxy.
    """
    implements(IZCache)
    security = ClassSecurityInfo()
    security.declareObjectPrivate()

    def __init__(self, proxy, request_names):
        self.proxy = proxy
        self.request_names = request_names

    def ZCache_invalidate(self, ob):
        """ See IZCache.
        """
        path = '/'.join(ob.getPhysicalPath())
        proxy = self.proxy
        keys = proxy.get(path)
        if keys is None:
            return
        for key in keys:
            proxy.remove(key)
        proxy.remove(path)

    def ZCache_get(self,
                   ob,
                   view_name='',
                   keywords=None,
                   mtime_func=None,
                   default=None,
                  ):
        """ See IZCache.
        """
        key = self._getKey(ob, view_name, keywords)

        value = self.proxy.get(key)

        if value is None:
            return default

        return value

    def ZCache_set(self,
                   ob,
                   data,
                   view_name='',
                   keywords=None,
                   mtime_func=None,
                  ):
        """ See IZCache.
        """
        path = '/'.join(ob.getPhysicalPath())
        proxy = self.proxy
        key = self._getKey(ob, view_name, keywords)
        proxy.set(key, data)

        keys = proxy.get(path)
        if keys is None:
            keys = {}
        if key not in keys:
            keys[key] = 1
            proxy.set(path, keys)

    def _getKey(self, ob, view_name, keywords):
        rnames = self.request_names
        if rnames:
            request = getattr(ob, 'REQUEST', {})
        else:
            request = {}

        return aggregateKey(ob, view_name, request, rnames, keywords)

InitializeClass(MemCacheZCache)

class MemCacheZCacheManager(CacheManager, SimpleItem, PropertyManager):
    """ Implement ISDC via a memcache proxy.
    """
    implements(IZCacheManager
             + implementedBy(CacheManager)
             + implementedBy(SimpleItem)
             + implementedBy(PropertyManager)
              )
    security = ClassSecurityInfo()

    _v_proxy = None
    proxy_path = None
    request_names = ()

    #
    #   ZMI
    #
    meta_type = 'MemCache Cache Manager'
    _properties = (
        {'id': 'proxy_path', 'type': 'string', 'mode': 'w'},
        {'id': 'request_names', 'type': 'lines', 'mode': 'w'},
    )

    manage_options = (PropertyManager.manage_options
                    + CacheManager.manage_options
                    + SimpleItem.manage_options
                     )

    security.declarePrivate('_get_proxy')
    def _get_proxy(self):
        if self._v_proxy is None:
            if self.proxy_path is None:
                # import late to avoid cycle
                from Products.mcdutils import MemCacheError
                raise MemCacheError('No proxy defined')
            self._v_proxy = self.unrestrictedTraverse(self.proxy_path)
        return self._v_proxy

    def ZCacheManager_getCache(self):
        """ See IZCacheManager.
        """
        names = list(self.request_names)
        names.sort()
        return MemCacheZCache(self._get_proxy(), tuple(names))

InitializeClass(MemCacheZCacheManager)


def addMemCacheZCacheManager(dispatcher, id, REQUEST):
    """ Add a MCSDC to dispatcher.
    """
    zcm = MemCacheZCacheManager()
    zcm._setId(id)
    dispatcher._setObject(id, zcm)
    REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                    % dispatcher.absolute_url())

addMemCacheZCacheManagerForm = PageTemplateFile('www/add_mczcm.pt', globals())



""" memcache-aware transactional mapping.

$Id: mapping.py,v 1.5 2006/06/29 21:13:36 tseaver Exp $
"""
from persistent.mapping import PersistentMapping
import transaction
from transaction.interfaces import IDataManager
from zope.interface import implementedBy
from zope.interface import implements

from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import Explicit
from Globals import InitializeClass

class MemCacheMapping(Explicit, PersistentMapping):
    """ memcache-based mapping which manages its own transactional semantics
    """
    implements(IDataManager + implementedBy(PersistentMapping))
    security = ClassSecurityInfo()

    def __init__(self, key, proxy):
        PersistentMapping.__init__(self)
        self._p_oid = hash(key)
        self._p_jar = self   # we are our own data manager
        self._p_key = key
        self._p_proxy = proxy
        self._p_joined = False

    security.setDefaultAccess('allow')
    security.declareObjectPublic()

    set = PersistentMapping.__setitem__
    __guarded_setitem__ = PersistentMapping.__setitem__
    __guarded_delitem__ = PersistentMapping.__delitem__
    delete = PersistentMapping.__delitem__

    def __getstate__(self):
        return self.data

    def __setstate__(self, value):
        self.data = {}
        self.data.update(value)

    def getContainerKey(self):
        """ Fake out (I)Transient API.
        """
        return self._p_key

    def _clean(self):
        # Remove from proxy cache to force an update
        # from memcached during next access.
        try:
            del self._p_proxy._cached[self._p_key]
        except KeyError:
            pass

    security.declarePrivate('abort')
    def abort(self, txn):
        """ See IDataManager.
        """
        self._clean()

    security.declarePrivate('tpc_begin')
    def tpc_begin(self, txn):
        """ See IDataManager.
        """

    security.declarePrivate('commit')
    def commit(self, txn):
        """ See IDataManager.
        """

    security.declarePrivate('tpc_vote')
    def tpc_vote(self, txn):
        """ See IDataManager.
        """
        server, key = self._p_proxy.client._get_server(self._p_key)
        if server is None:
            from Products.mcdutils import MemCacheError
            raise MemCacheError("Can't reach memcache server!")

    security.declarePrivate('tpc_finish')
    def tpc_finish(self, txn):
        """ See IDataManager.
        """
        if self._p_changed:
            self._p_proxy.set(self._p_key, self) # no error handling
        self._p_changed = 0
        self._p_joined = False
        self._clean()

    security.declarePrivate('tpc_abort')
    def tpc_abort(self, txn):
        """ See IDataManager.
        """
        self._p_joined = False
        self._p_changed = 0
        self._clean()

    security.declarePrivate('sortKey')
    def sortKey(self):
        """ See IDataManager.
        """
        return 'MemCacheMapping: %s' % self._p_key

    security.declarePrivate('register')
    def register(self, obj):
        """ See IPersistentDataManager
        """
        if obj is not self:
            raise ValueError, "Can't be the jar for another object."

        if not self._p_joined:
            transaction.get().join(self)
            self._p_joined = True

InitializeClass(MemCacheMapping)



""" mcdutils product interfaces

$Id: interfaces.py,v 1.6 2006/06/07 05:43:09 tseaver Exp $
"""
from transaction.interfaces import IDataManager
from zope.interface import Attribute
from zope.interface import Interface

class ISessionDataContainer(Interface):
    """ Document the implied interface expected by Zope2's SessionDataManager.
    """
    def has_key(key):
        """ Return True if the container has the key, else False.
        """

    def new_or_existing(key):
        """ Return a mapping for 'key', creating it if needed.

        o The returned object must be Acquisition-wrappable.
        """

    def get(key):
        """ Return a mapping for 'key'.

        o Return None of no mapping exists.

        o If not None, the returned object must be Acquisition-wrappable.
        """

class IMemCacheMapping(IDataManager):
    """ Combine Python's mapping protocol with transaction management.
    """

class IMemCacheProxy(Interface):
    """ Manage client connection to a pool of memcached servers.
    """
    servers = Attribute(u'servers', u"""List of servers

Each item is a <host>:<port> server address.""")

    client = Attribute(u'client', u"""memcache.Client instance"""
                      )
    client.setTaggedValue('read_only', True)

    def get(key):
        """ Return the value stored in the cache under 'key'.
        """

    def get_multi(keys):
        """ Return a mapping of values stored in the cache under 'keys'.
        """

    def set(key, value):
        """ Store value for 'key'.

        o Return a boolean to indicate success.
        """

    def add(key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key does not already exist.
        """

    def replace(key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key already exists.
        """

    def delete(key, time=0):
        """ Remove the value stored in the cache under 'key'.

        o 'time', if passed an integer time value (in seconds) during
          which the memcached server will block new writes to this key
          (perhaps useful in preventing race conditions).

        o Return a boolean to indicate success.
        """

class IMemCacheSessionDataContainer(ISessionDataContainer):
    """ memcache-specific SDC, using a proxy.
    """
    proxy_path = Attribute(u"""Path to proxy.

No session operations are possible if the path is invalid.""")


class IZCache(Interface):
    """ Interface describing API for OFS.Cache.Cache.
    """
    def ZCache_invalidate(ob):
        """ Remove any entries from the cache for 'ob'.
        """

    def ZCache_get(ob, view_name, keywords, mtime_func, default):
        """ Fetch a cache entry for 'ob'.

        o If an object provides different views that would benefit from
          caching, it will set 'view_name', which should be treated as
          part of the cache key.  It defaults to the empty string.
         
        o 'keywords', if passed, will be a mapping containing keys that
          distinguish this cache entry from others even though
          'ob' and 'view_name' are the same;  the value should thus be
          part of the key for the entry.  DTMLMethods use keywords
          derived from the DTML namespace.
         
        o When the Cache calls 'ob.ZCacheable_getModTime',
          it should pass 'mtime_func' as an argument.  It is provided to
          allow cacheable objects to provide their own computation
          of the object's modification time.
         
        o If no matching entry is found, return 'default'.
        """

    def ZCache_set(ob, data, view_name, keywords, mtime_func):
        """ Store a value in the cache for 'ob'.

        o 'data' is the value to be stored.

        o See ZCache_get() for description of the 'keywords', 'mtime_func',
          and 'default' parameters.
        """

class IZCacheManager(Interface):
    """ Interface describing API for OFS.Cache.CacheManager. 
    """
    def ZCacheManager_getCache():
        """ Return an object implementing IZCache.
        """



""" Unit tests for Products.mcdutils.proxy

$Id: test_sessiondata.py,v 1.5 2006/10/05 21:04:25 tseaver Exp $
"""
import unittest

class DummyClient:
    def _get_server(self, key):
        return self, key

class DummyProxy:
    def __init__(self):
        self._cached = {}

    def set(self, key, value):
        pass

    def _get(self, key, default=None):
        return self._cached.get(key, default)

    get = _get

class MemCacheSessionDataTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.sessiondata \
            import MemCacheSessionDataContainer
        return MemCacheSessionDataContainer

    def _makeOne(self, with_proxy=True, *args, **kw):
        sdc = self._getTargetClass()(*args, **kw)
        if with_proxy:
            sdc.dummy_proxy = DummyProxy()
            sdc.proxy_path = 'dummy_proxy'
        return sdc

    def test_conforms_to_ISessionDataContainer(self):
        from zope.interface.verify import verifyClass
        from Products.mcdutils.interfaces import ISessionDataContainer
        verifyClass(ISessionDataContainer, self._getTargetClass())

    def test_conforms_to_IMemCacheSessionDataContainer(self):
        from zope.interface.verify import verifyClass
        from Products.mcdutils.interfaces import IMemCacheSessionDataContainer
        verifyClass(IMemCacheSessionDataContainer, self._getTargetClass())

    def test_empty(self):
        sdc = self._makeOne()
        self.assertEqual(sdc.has_key('foobar'), False)
        self.assertEqual(sdc.get('foobar'), None)

    def test_invalid_proxy_raises_MemCacheError(self):
        from Products.mcdutils import MemCacheError
        sdc = self._makeOne(with_proxy=False)
        self.assertRaises(MemCacheError, sdc.has_key, 'foobar')
        self.assertRaises(MemCacheError, sdc.get, 'foobar')
        self.assertRaises(MemCacheError, sdc.new_or_existing, 'foobar')

    def test_new_or_existing_returns_txn_aware_mapping(self):
        from persistent.mapping import PersistentMapping
        from transaction.interfaces import IDataManager
        sdc = self._makeOne()
        created = sdc.new_or_existing('foobar')
        self.failUnless(isinstance(created, PersistentMapping))
        jar = created._p_jar
        self.failIf(jar is None)
        self.failUnless(IDataManager.providedBy(jar))

    def test_has_key_after_new_or_existing_returns_True(self):
        from persistent.mapping import PersistentMapping
        sdc = self._makeOne()
        created = sdc.new_or_existing('foobar')
        self.assertEqual(sdc.has_key('foobar'), True)

    def test_get_after_new_or_existing_returns_same(self):
        sdc = self._makeOne()
        created = sdc.new_or_existing('foobar')
        self.failUnless(sdc.get('foobar') is created)

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheSessionDataTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



""" Unit tests for Products.mcdutils.proxy

$Id: test_proxy.py,v 1.7 2006/05/31 20:57:16 tseaver Exp $
"""
import unittest

class MemCacheProxyTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.proxy import MemCacheProxy
        return MemCacheProxy

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_conforms_to_IMemCacheProxy(self):
        from zope.interface.verify import verifyClass
        from Products.mcdutils.interfaces import IMemCacheProxy
        verifyClass(IMemCacheProxy, self._getTargetClass())

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheProxyTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# Run this test from 'zopectl run'
# Requires that we are running a memcached on localhost, port 11211
import transaction
from Products.mcdutils.proxy import MemCacheProxy

proxy = MemCacheProxy(['localhost:11211'])

session = proxy.new_or_existing('foobar')
print session

session['abc'] = 123
print session

transaction.commit()

proxy2 = MemCacheProxy(['localhost:11211'])

print proxy2.get('foobar')



#!/usr/bin/env python

"""
client module for memcached (memory cache daemon)

Overview
========

See U{the MemCached homepage<http://www.danga.com/memcached>} for more about memcached.

Usage summary
=============

This should give you a feel for how this module operates::

    import memcache
    mc = memcache.Client(['127.0.0.1:11211'], debug=0)

    mc.set("some_key", "Some value")
    value = mc.get("some_key")

    mc.set("another_key", 3)
    mc.delete("another_key")
    
    mc.set("key", "1")   # note that the key used for incr/decr must be a string.
    mc.incr("key")
    mc.decr("key")

The standard way to use memcache with a database is like this::

    key = derive_key(obj)
    obj = mc.get(key)
    if not obj:
        obj = backend_api.get(...)
        mc.set(obj)

    # we now have obj, and future passes through this code
    # will use the object from the cache.

Detailed Documentation
======================

More detailed documentation is available in the L{Client} class.
"""

import sys
import socket
import time
import types
try:
    import cPickle as pickle
except ImportError:
    import pickle

__author__    = "Evan Martin <martine@danga.com>"
__version__   = "1.2_tummy6"
__copyright__ = "Copyright (C) 2003 Danga Interactive"
__license__   = "Python"

class _Error(Exception):
    pass

class Client:
    """
    Object representing a pool of memcache servers.
    
    See L{memcache} for an overview.

    In all cases where a key is used, the key can be either:
        1. A simple hashable type (string, integer, etc.).
        2. A tuple of C{(hashvalue, key)}.  This is useful if you want to avoid
        making this module calculate a hash value.  You may prefer, for
        example, to keep all of a given user's objects on the same memcache
        server, so you could use the user's unique id as the hash value.

    @group Setup: __init__, set_servers, forget_dead_hosts, disconnect_all, debuglog
    @group Insertion: set, add, replace
    @group Retrieval: get, get_multi
    @group Integers: incr, decr
    @group Removal: delete
    @sort: __init__, set_servers, forget_dead_hosts, disconnect_all, debuglog,\
           set, add, replace, get, get_multi, incr, decr, delete
    """
    _FLAG_PICKLE  = 1<<0
    _FLAG_INTEGER = 1<<1
    _FLAG_LONG    = 1<<2

    _SERVER_RETRIES = 10  # how many times to try finding a free server.

    def __init__(self, servers, debug=0):
        """
        Create a new Client object with the given list of servers.

        @param servers: C{servers} is passed to L{set_servers}.
        @param debug: whether to display error messages when a server can't be
        contacted.
        """
        self.set_servers(servers)
        self.debug = debug
        self.stats = {}
    
    def set_servers(self, servers):
        """
        Set the pool of servers used by this client.

        @param servers: an array of servers.
        Servers can be passed in two forms:
            1. Strings of the form C{"host:port"}, which implies a default weight of 1.
            2. Tuples of the form C{("host:port", weight)}, where C{weight} is
            an integer weight value.
        """
        self.servers = [_Host(s, self.debuglog) for s in servers]
        self._init_buckets()

    def get_stats(self):
        '''Get statistics from each of the servers.  

        @return: A list of tuples ( server_identifier, stats_dictionary ).
            The dictionary contains a number of name/value pairs specifying
            the name of the status field and the string value associated with
            it.  The values are not converted from strings.
        '''
        data = []
        for s in self.servers:
            if not s.connect(): continue
            name = '%s:%s (%s)' % ( s.ip, s.port, s.weight )
            s.send_cmd('stats')
            serverData = {}
            data.append(( name, serverData ))
            readline = s.readline
            while 1:
                line = readline()
                if not line or line.strip() == 'END': break
                stats = line.split(' ', 2)
                serverData[stats[1]] = stats[2]

        return(data)

    def flush_all(self):
        'Expire all data currently in the memcache servers.'
        for s in self.servers:
            if not s.connect(): continue
            s.send_cmd('flush_all')
            s.expect("OK")

    def debuglog(self, str):
        if self.debug:
            sys.stderr.write("MemCached: %s\n" % str)

    def _statlog(self, func):
        if not self.stats.has_key(func):
            self.stats[func] = 1
        else:
            self.stats[func] += 1

    def forget_dead_hosts(self):
        """
        Reset every host in the pool to an "alive" state.
        """
        for s in self.servers:
            s.dead_until = 0

    def _init_buckets(self):
        self.buckets = []
        for server in self.servers:
            for i in range(server.weight):
                self.buckets.append(server)

    def _get_server(self, key):
        if type(key) == types.TupleType:
            serverhash = key[0]
            key = key[1]
        else:
            serverhash = hash(key)

        for i in range(Client._SERVER_RETRIES):
            server = self.buckets[serverhash % len(self.buckets)]
            if server.connect():
                #print "(using server %s)" % server,
                return server, key
            serverhash = hash(str(serverhash) + str(i))
        return None, None

    def disconnect_all(self):
        for s in self.servers:
            s.close_socket()
    
    def delete(self, key, time=0):
        '''Deletes a key from the memcache.
        
        @return: Nonzero on success.
        @rtype: int
        '''
        server, key = self._get_server(key)
        if not server:
            return 0
        self._statlog('delete')
        if time != None:
            cmd = "delete %s %d" % (key, time)
        else:
            cmd = "delete %s" % key

        try:
            server.send_cmd(cmd)
            server.expect("DELETED")
        except socket.error, msg:
            server.mark_dead(msg[1])
            return 0
        return 1

    def incr(self, key, delta=1):
        """
        Sends a command to the server to atomically increment the value for C{key} by
        C{delta}, or by 1 if C{delta} is unspecified.  Returns None if C{key} doesn't
        exist on server, otherwise it returns the new value after incrementing.

        Note that the value for C{key} must already exist in the memcache, and it
        must be the string representation of an integer.

        >>> mc.set("counter", "20")  # returns 1, indicating success
        1
        >>> mc.incr("counter")
        21
        >>> mc.incr("counter")
        22

        Overflow on server is not checked.  Be aware of values approaching
        2**32.  See L{decr}.

        @param delta: Integer amount to increment by (should be zero or greater).
        @return: New value after incrementing.
        @rtype: int
        """
        return self._incrdecr("incr", key, delta)

    def decr(self, key, delta=1):
        """
        Like L{incr}, but decrements.  Unlike L{incr}, underflow is checked and
        new values are capped at 0.  If server value is 1, a decrement of 2
        returns 0, not -1.

        @param delta: Integer amount to decrement by (should be zero or greater).
        @return: New value after decrementing.
        @rtype: int
        """
        return self._incrdecr("decr", key, delta)

    def _incrdecr(self, cmd, key, delta):
        server, key = self._get_server(key)
        if not server:
            return 0
        self._statlog(cmd)
        cmd = "%s %s %d" % (cmd, key, delta)
        try:
            server.send_cmd(cmd)
            line = server.readline()
            return int(line)
        except socket.error, msg:
            server.mark_dead(msg[1])
            return None

    def add(self, key, val, time=0):
        '''
        Add new key with value.
        
        Like L{set}, but only stores in memcache if the key doesn't already exist.

        @return: Nonzero on success.
        @rtype: int
        '''
        return self._set("add", key, val, time)
    def replace(self, key, val, time=0):
        '''Replace existing key with value.
        
        Like L{set}, but only stores in memcache if the key already exists.  
        The opposite of L{add}.

        @return: Nonzero on success.
        @rtype: int
        '''
        return self._set("replace", key, val, time)
    def set(self, key, val, time=0):
        '''Unconditionally sets a key to a given value in the memcache.

        The C{key} can optionally be an tuple, with the first element being the
        hash value, if you want to avoid making this module calculate a hash value.
        You may prefer, for example, to keep all of a given user's objects on the
        same memcache server, so you could use the user's unique id as the hash
        value.

        @return: Nonzero on success.
        @rtype: int
        '''
        return self._set("set", key, val, time)
    
    def _set(self, cmd, key, val, time):
        server, key = self._get_server(key)
        if not server:
            return 0

        self._statlog(cmd)

        flags = 0
        if isinstance(val, types.StringTypes):
            pass
        elif isinstance(val, int):
            flags |= Client._FLAG_INTEGER
            val = "%d" % val
        elif isinstance(val, long):
            flags |= Client._FLAG_LONG
            val = "%d" % val
        else:
            flags |= Client._FLAG_PICKLE
            val = pickle.dumps(val, 2)
        
        fullcmd = "%s %s %d %d %d\r\n%s" % (cmd, key, flags, time, len(val), val)
        try:
            server.send_cmd(fullcmd)
            server.expect("STORED")
        except socket.error, msg:
            server.mark_dead(msg[1])
            return 0
        return 1

    def get(self, key):
        '''Retrieves a key from the memcache.
        
        @return: The value or None.
        '''
        server, key = self._get_server(key)
        if not server:
            return None

        self._statlog('get')

        try:
            server.send_cmd("get %s" % key)
            rkey, flags, rlen, = self._expectvalue(server)
            if not rkey:
                return None
            value = self._recv_value(server, flags, rlen)
            server.expect("END")
        except (_Error, socket.error), msg:
            if type(msg) is types.TupleType:
                msg = msg[1]
            server.mark_dead(msg)
            return None
        return value

    def get_multi(self, keys):
        '''
        Retrieves multiple keys from the memcache doing just one query.
        
        >>> success = mc.set("foo", "bar")
        >>> success = mc.set("baz", 42)
        >>> mc.get_multi(["foo", "baz", "foobar"]) == {"foo": "bar", "baz": 42}
        1

        This method is recommended over regular L{get} as it lowers the number of
        total packets flying around your network, reducing total latency, since
        your app doesn't have to wait for each round-trip of L{get} before sending
        the next one.

        @param keys: An array of keys.
        @return:  A dictionary of key/value pairs that were available.

        '''

        self._statlog('get_multi')

        server_keys = {}

        # build up a list for each server of all the keys we want.
        for key in keys:
            server, key = self._get_server(key)
            if not server:
                continue
            if not server_keys.has_key(server):
                server_keys[server] = []
            server_keys[server].append(key)

        # send out all requests on each server before reading anything
        dead_servers = []
        for server in server_keys.keys():
            try:
                server.send_cmd("get %s" % " ".join(server_keys[server]))
            except socket.error, msg:
                server.mark_dead(msg[1])
                dead_servers.append(server)

        # if any servers died on the way, don't expect them to respond.
        for server in dead_servers:
            del server_keys[server]

        retvals = {}
        for server in server_keys.keys():
            try:
                line = server.readline()
                while line and line != 'END':
                    rkey, flags, rlen = self._expectvalue(server, line)
                    #  Bo Yang reports that this can sometimes be None
                    if rkey is not None:
                        val = self._recv_value(server, flags, rlen)
                        retvals[rkey] = val
                    line = server.readline()
            except (_Error, socket.error), msg:
                server.mark_dead(msg)
        return retvals

    def _expectvalue(self, server, line=None):
        if not line:
            line = server.readline()

        if line[:5] == 'VALUE':
            resp, rkey, flags, len = line.split()
            flags = int(flags)
            rlen = int(len)
            return (rkey, flags, rlen)
        else:
            return (None, None, None)

    def _recv_value(self, server, flags, rlen):
        rlen += 2 # include \r\n
        buf = server.recv(rlen)
        if len(buf) != rlen:
            raise _Error("received %d bytes when expecting %d" % (len(buf), rlen))

        if len(buf) == rlen:
            buf = buf[:-2]  # strip \r\n

        if flags == 0:
            val = buf
        elif flags & Client._FLAG_INTEGER:
            val = int(buf)
        elif flags & Client._FLAG_LONG:
            val = long(buf)
        elif flags & Client._FLAG_PICKLE:
            try:
                val = pickle.loads(buf)
            except:
                self.debuglog('Pickle error...\n')
                val = None
        else:
            self.debuglog("unknown flags on get: %x\n" % flags)

        return val

    
class _Host:
    _DEAD_RETRY = 30  # number of seconds before retrying a dead server.

    def __init__(self, host, debugfunc=None):
        if isinstance(host, types.TupleType):
            host = host[0]
            self.weight = host[1]
        else:
            self.weight = 1

        if host.find(":") > 0:
            self.ip, self.port = host.split(":")
            self.port = int(self.port)
        else:
            self.ip, self.port = host, 11211

        if not debugfunc:
            debugfunc = lambda x: x
        self.debuglog = debugfunc

        self.deaduntil = 0
        self.socket = None
    
    def _check_dead(self):
        if self.deaduntil and self.deaduntil > time.time():
            return 1
        self.deaduntil = 0
        return 0

    def connect(self):
        if self._get_socket():
            return 1
        return 0

    def mark_dead(self, reason):
        self.debuglog("MemCache: %s: %s.  Marking dead." % (self, reason))
        self.deaduntil = time.time() + _Host._DEAD_RETRY
        self.close_socket()
        
    def _get_socket(self):
        if self._check_dead():
            return None
        if self.socket:
            return self.socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Python 2.3-ism:  s.settimeout(1)
        try:
            s.connect((self.ip, self.port))
        except socket.error, msg:
            self.mark_dead("connect: %s" % msg[1])
            return None
        self.socket = s
        return s
    
    def close_socket(self):
        if self.socket:
            self.socket.close()
            self.socket = None

    def send_cmd(self, cmd):
        self.socket.sendall(cmd + '\r\n')

    def readline(self):
        buffers = ''
        recv = self.socket.recv
        while 1:
            data = recv(1)
            if not data:
                self.mark_dead('Connection closed while reading from %s'
                        % repr(self))
                break
            if data == '\n' and buffers and buffers[-1] == '\r':
                return(buffers[:-1])
            buffers = buffers + data
        return(buffers)

    def expect(self, text):
        line = self.readline()
        if line != text:
            self.debuglog("while expecting '%s', got unexpected response '%s'" % (text, line))
        return line
    
    def recv(self, rlen):
        buf = ''
        recv = self.socket.recv
        while len(buf) < rlen:
            buf = buf + recv(rlen - len(buf))
        return buf

    def __str__(self):
        d = ''
        if self.deaduntil:
            d = " (dead until %d)" % self.deaduntil
        return "%s:%d%s" % (self.ip, self.port, d)

def _doctest():
    import doctest, memcache
    servers = ["127.0.0.1:11211"]
    mc = Client(servers, debug=1)
    globs = {"mc": mc}
    return doctest.testmod(memcache, globs=globs)

if __name__ == "__main__":
    print "Testing docstrings..."
    _doctest()
    print "Running tests:"
    print
    #servers = ["127.0.0.1:11211", "127.0.0.1:11212"]
    servers = ["127.0.0.1:11211"]
    mc = Client(servers, debug=1)

    def to_s(val):
        if not isinstance(val, types.StringTypes):
            return "%s (%s)" % (val, type(val))
        return "%s" % val
    def test_setget(key, val):
        print "Testing set/get {'%s': %s} ..." % (to_s(key), to_s(val)),
        mc.set(key, val)
        newval = mc.get(key)
        if newval == val:
            print "OK"
            return 1
        else:
            print "FAIL"
            return 0

    class FooStruct:
        def __init__(self):
            self.bar = "baz"
        def __str__(self):
            return "A FooStruct"
        def __eq__(self, other):
            if isinstance(other, FooStruct):
                return self.bar == other.bar
            return 0
        
    test_setget("a_string", "some random string")
    test_setget("an_integer", 42)
    if test_setget("long", long(1<<30)):
        print "Testing delete ...",
        if mc.delete("long"):
            print "OK"
        else:
            print "FAIL"
    print "Testing get_multi ...",
    print mc.get_multi(["a_string", "an_integer"])

    print "Testing get(unknown value) ...",
    print to_s(mc.get("unknown_value"))

    f = FooStruct()
    test_setget("foostruct", f)

    print "Testing incr ...",
    x = mc.incr("an_integer", 1)
    if x == 43:
        print "OK"
    else:
        print "FAIL"

    print "Testing decr ...",
    x = mc.decr("an_integer", 1)
    if x == 42:
        print "OK"
    else:
        print "FAIL"



# vim: ts=4 sw=4 et :



""" mcdutils session data container

$Id: sessiondata.py,v 1.3 2006/06/06 22:54:05 tseaver Exp $
"""
from zope.interface import implementedBy
from zope.interface import implements

from AccessControl.SecurityInfo import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.mcdutils.interfaces import IMemCacheSessionDataContainer
from Products.mcdutils.mapping import MemCacheMapping

class MemCacheSessionDataContainer(SimpleItem, PropertyManager):
    """ Implement ISDC via a memcache proxy.
    """
    implements(IMemCacheSessionDataContainer
             + implementedBy(SimpleItem)
             + implementedBy(PropertyManager)
              )
    security = ClassSecurityInfo()

    _v_proxy = None
    proxy_path = None

    security.declarePrivate('_get_proxy')
    def _get_proxy(self):
        if self._v_proxy is None:
            if self.proxy_path is None:
                from Products.mcdutils import MemCacheError
                raise MemCacheError('No proxy defined')
            self._v_proxy = self.unrestrictedTraverse(self.proxy_path)
        return self._v_proxy

    #proxy = property(_get_proxy,)  # XXX can't acquire inside property!

    #
    #   ZMI
    #
    meta_type = 'MemCache Session Data Container'
    _properties = (
        {'id': 'proxy_path', 'type': 'string', 'mode': 'w'},
    )

    manage_options = (PropertyManager.manage_options
                    + ({'action': 'addItemsToSessionForm', 'label': 'Test'},)
                    + SimpleItem.manage_options
                     )

    security.declarePublic('addItemsToSessionForm')
    addItemsToSessionForm = PageTemplateFile('www/add_items.pt', globals())

    security.declarePublic('addItemsToSession')
    def addItemsToSession(self):
        """ Add key value pairs from 'items' textarea to the session.
        """
        request = self.REQUEST
        items = request.form.get('items', ())
        session = request['SESSION']

        before = len(session)
        count = len(items)

        for line in items:
            k, v = line.split(' ', 1)
            k = k.strip()
            v = v.strip()
            session[k] = v

        after = len(session)

        return 'Before: %d;  after: %d; # items: %d' % (before, after, count)

    #
    #   ISessionDataContainer implementation
    #
    security.declarePrivate('has_key')
    def has_key(self, key):
        """ See ISessionDataContainer.
        """
        return self._get_proxy().get(key) is not None

    security.declarePrivate('new_or_existing')
    def new_or_existing(self, key):
        """ See ISessionDataContainer.
        """
        mapping = self.get(key)

        if mapping is None:
            proxy = self._get_proxy()
            mapping = MemCacheMapping(key, proxy)
            proxy._cached[key] = mapping # XXX

        return mapping

    security.declarePrivate('get')
    def get(self, key):
        """ See ISessionDataContainer.
        """
        return self._get_proxy().get(key)

InitializeClass(MemCacheSessionDataContainer)


def addMemCacheSessionDataContainer(dispatcher, id, REQUEST):
    """ Add a MCSDC to dispatcher.
    """
    sdc = MemCacheSessionDataContainer()
    sdc._setId(id)
    dispatcher._setObject(id, sdc)
    REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                    % dispatcher.absolute_url())

addMemCacheSessionDataContainerForm = PageTemplateFile('www/add_mcsdc.pt',
                                                       globals())



""" Unit tests for Products.mcdutils.zcache

$Id: test_zcache.py,v 1.1 2006/06/07 05:43:09 tseaver Exp $
"""
import unittest

class TestsOf_aggregateKey(unittest.TestCase):

    def test_defaults(self):
        from Products.mcdutils.zcache import aggregateKey
        key = aggregateKey(DummyOb())
        self.assertEqual(key, '%s|||' % _DUMMY_PATH_STR)

    def test_explicit_view_name(self):
        from Products.mcdutils.zcache import aggregateKey
        key = aggregateKey(DummyOb(), view_name='VIEW_NAME')
        self.assertEqual(key, '%s|VIEW_NAME||' % _DUMMY_PATH_STR)

    def test_explicit_request_names(self):
        from Products.mcdutils.zcache import aggregateKey
        key = aggregateKey(DummyOb(),
                           request={'aaa': 'AAA',
                                    'bbb': 'BBB',
                                    'ccc': 'CCC',
                                   },
                           request_names=['aaa', 'ccc'],
                          )
        self.assertEqual(key, '%s||aaa:AAA,ccc:CCC|' % _DUMMY_PATH_STR)

    def test_explicit_local_keys(self):
        from Products.mcdutils.zcache import aggregateKey
        key = aggregateKey(DummyOb(), local_keys={'foo': 'bar', 'baz': 'bam'})
        self.assertEqual(key, '%s|||baz:bam,foo:bar' % _DUMMY_PATH_STR)

class MemCacheZCacheTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.zcache import MemCacheZCache
        return MemCacheZCache

    def _makeOne(self, proxy, request_names=(), *args, **kw):
        mczc = self._getTargetClass()(proxy, request_names, *args, **kw)
        return mczc

    def test_conforms_to_IZCache(self):
        from zope.interface.verify import verifyClass
        from Products.mcdutils.interfaces import IZCache

        verifyClass(IZCache, self._getTargetClass())

    def test_ZCache_get_cache_miss(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        self.assertEqual(cache.ZCache_get(DummyOb()), None)

    def test_ZCache_get_cache_hit_default_args(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        proxy._cached['%s|||' % _DUMMY_PATH_STR] = 'XYZZY'

        self.assertEqual(cache.ZCache_get(DummyOb()), 'XYZZY')

    def test_ZCache_get_cache_hit_view_name(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 
        proxy._cached['%s|||' % _DUMMY_PATH_STR] = 'XYZZY'
        proxy._cached['%s|foo||' % _DUMMY_PATH_STR] = 'ABCDEF'

        self.assertEqual(cache.ZCache_get(DummyOb(), view_name='foo'),
                         'ABCDEF')

    def test_ZCache_get_cache_miss_view_name(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 
        proxy._cached['%s|||' % _DUMMY_PATH_STR] = 'XYZZY'
        proxy._cached['%s|foo||' % _DUMMY_PATH_STR] = 'ABCDEF'

        self.assertEqual(cache.ZCache_get(DummyOb(), view_name='bar'), None)

    def test_ZCache_get_cache_hit_request_names(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy, request_names=('bar', 'qux'))

        _cached = proxy._cached 
        proxy._cached['%s|||' % _DUMMY_PATH_STR] = 'XYZZY'
        proxy._cached['%s||bar:baz,qux:|' % _DUMMY_PATH_STR] = 'ABCDEF'

        ob = DummyOb()
        ob.REQUEST = {'bar': 'baz', 'bam': 'bif'}

        self.assertEqual(cache.ZCache_get(ob), 'ABCDEF')

    def test_ZCache_invalidate(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 
        proxy._cached['%s|||' % _DUMMY_PATH_STR] = 'XYZZY'
        proxy._cached['%s|foo||' % _DUMMY_PATH_STR] = 'ABCDEF'
        proxy._cached['%s|bar||' % _DUMMY_PATH_STR] = 'LMNOP'

        keys = _cached.keys()
        _cached[_DUMMY_PATH_STR] = dict([(k, 1) for k in keys])

        cache.ZCache_invalidate(DummyOb())

        self.assertEqual(len(_cached), 0)

    def test_ZCache_set_simple(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 

        cache.ZCache_set(DummyOb(), 'XYZZY')

        self.assertEqual(len(_cached), 2)
        key = '%s|||' % _DUMMY_PATH_STR
        self.failUnless(key in _cached[_DUMMY_PATH_STR])
        self.assertEqual(_cached[key], 'XYZZY')

    def test_ZCache_set_with_view_name(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 

        cache.ZCache_set(DummyOb(), 'XYZZY', view_name='v')

        self.assertEqual(len(_cached), 2)
        key = '%s|v||' % _DUMMY_PATH_STR
        self.failUnless(key in _cached[_DUMMY_PATH_STR])
        self.assertEqual(_cached[key], 'XYZZY')

    def test_ZCache_set_replacing(self):
        proxy = DummyProxy()
        cache = self._makeOne(proxy)

        _cached = proxy._cached 
        key1 = '%s|||' % _DUMMY_PATH_STR
        key2 = '%s|v||' % _DUMMY_PATH_STR
        _cached[_DUMMY_PATH_STR] = {key1: 1, key2: 1}
        _cached[key1] = 'GHIJKL'
        _cached[key2] = 'ABCDE'

        cache.ZCache_set(DummyOb(), 'XYZZY', view_name='v')

        self.assertEqual(len(_cached), 3)

        self.failUnless(key1 in _cached[_DUMMY_PATH_STR])
        self.assertEqual(_cached[key1], 'GHIJKL')

        self.failUnless(key2 in _cached[_DUMMY_PATH_STR])
        self.assertEqual(_cached[key2], 'XYZZY')

class MemCacheZCacheManagerTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.zcache import MemCacheZCacheManager
        return MemCacheZCacheManager

    def _makeOne(self, *args, **kw):
        mczc = self._getTargetClass()(*args, **kw)
        return mczc

    def test_conforms_to_IZCacheManager(self):
        from zope.interface.verify import verifyClass
        from Products.mcdutils.interfaces import IZCacheManager

        verifyClass(IZCacheManager, self._getTargetClass())

    def test_ZCacheManager_getCache_with_proxy(self):
        mgr = self._makeOne()
        mgr.dummy_proxy = DummyProxy()
        mgr.proxy_path = 'dummy_proxy'
        mgr.request_names = ('foo', 'bar')

        cache = mgr.ZCacheManager_getCache()

        self.assertEqual(cache.proxy, mgr.dummy_proxy)
        self.assertEqual(cache.request_names, ('bar', 'foo'))


_DUMMY_PATH = ('path', 'to', 'dummy')
_DUMMY_PATH_STR = '/'.join(_DUMMY_PATH)

class DummyOb:
    def getPhysicalPath(self):
        return _DUMMY_PATH

class DummyProxy:
    def __init__(self):
        self._cached = {}

    def set(self, key, value):
        self._cached[key] = value

    def _get(self, key, default=None):
        return self._cached.get(key, default)

    get = _get

    def remove(self, key):
        del self._cached[key]

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestsOf_aggregateKey))
    suite.addTest(unittest.makeSuite(MemCacheZCacheTests))
    suite.addTest(unittest.makeSuite(MemCacheZCacheManagerTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



""" Unit tests for Products.mcdutils.mapping

$Id: test_mapping.py,v 1.2 2006/05/31 20:57:16 tseaver Exp $
"""
import unittest

class MemCacheMappingTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.mapping import MemCacheMapping
        return MemCacheMapping

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_conforms_to_IDataManager(self):
        from zope.interface.verify import verifyClass
        from transaction.interfaces import IDataManager
        verifyClass(IDataManager, self._getTargetClass())

    def test___setitem___triggers_register(self):
        mapping = self._makeOne('key', DummyProxy())
        self.failIf(mapping._p_changed)
        self.failIf(mapping._p_joined)
        mapping['abc'] = 123
        self.failUnless(mapping._p_changed)
        self.failUnless(mapping._p_joined)

class DummyClient:
    def _get_server(self, key):
        return self, key

class DummyProxy:
    def _set(self, key, value):
        pass

    client = DummyClient()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheMappingTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



""" Product:  mcdutils

Implement Zope sessions using memcached as the backing store.

$Id: __init__.py,v 1.6 2006/06/07 05:43:09 tseaver Exp $
"""

class MemCacheError(IOError):
    pass

def initialize(context):

    from Products.mcdutils.proxy import MemCacheProxy
    from Products.mcdutils.proxy import addMemCacheProxyForm
    from Products.mcdutils.proxy import addMemCacheProxy
    context.registerClass(MemCacheProxy,
                          constructors=(addMemCacheProxyForm,
                                        addMemCacheProxy),
                          icon='www/proxy.gif',
                         )

    from Products.mcdutils.sessiondata import MemCacheSessionDataContainer
    from Products.mcdutils.sessiondata \
        import addMemCacheSessionDataContainerForm
    from Products.mcdutils.sessiondata import addMemCacheSessionDataContainer
    context.registerClass(MemCacheSessionDataContainer,
                          constructors=(addMemCacheSessionDataContainerForm,
                                        addMemCacheSessionDataContainer),
                          icon='www/sdc.gif',
                         )

    from Products.mcdutils.zcache import MemCacheZCacheManager
    from Products.mcdutils.zcache \
        import addMemCacheZCacheManagerForm
    from Products.mcdutils.zcache import addMemCacheZCacheManager
    context.registerClass(MemCacheZCacheManager,
                          constructors=(addMemCacheZCacheManagerForm,
                                        addMemCacheZCacheManager),
                          icon='www/zcm.gif',
                         )



""" Functional tests for Products.mcdutils.proxy
"""

import unittest

class MemCacheSDCFuncTests(unittest.TestCase):

    def _makeOne(self):
        from Products.mcdutils.sessiondata import MemCacheSessionDataContainer
        from Products.mcdutils.proxy import MemCacheProxy
        sdc = MemCacheSessionDataContainer()
        sdc.mcproxy = MemCacheProxy()
        sdc.proxy_path = 'mcproxy'

        return sdc

    def test_writing_to_mapping_no_memcache(self):
        from Products.mcdutils.mapping import MemCacheMapping

        sdc = self._makeOne()
        mapping = sdc.new_or_existing('foobar')
        self.failUnless(isinstance(mapping, MemCacheMapping))
        self.failIf(mapping._p_changed)
        self.failIf(mapping._p_joined)
        mapping['abc'] = 1345
        self.failUnless(mapping._p_changed)
        self.failUnless(mapping._p_joined)
        import transaction
        transaction.commit()

    def test_writing_to_mapping_with_memcache(self):
        from Products.mcdutils.mapping import MemCacheMapping

        sdc = self._makeOne()
        sdc._get_proxy().servers = ('localhost:11211',)
        mapping = sdc.new_or_existing('foobar')
        self.failUnless(isinstance(mapping, MemCacheMapping))
        self.failIf(mapping._p_changed)
        self.failUnless(mapping._p_joined)
        mapping['abc'] = 1345
        self.failUnless(mapping._p_changed)
        self.failUnless(mapping._p_joined)
        import transaction
        transaction.commit()

    def test_writing_to_mapping_with_invalid_memcache_raises(self):
        from Products.mcdutils import MemCacheError

        sdc = self._makeOne()
        sdc._get_proxy().servers = ('nonesuch:999999',)
        mapping = sdc.new_or_existing('foobar')
        mapping['abc'] = 1345
        import transaction
        self.assertRaises(MemCacheError, transaction.commit)
        transaction.abort()

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheSDCFuncTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



from setuptools import setup, find_packages

readme = open('README.txt').read()
history = open('CHANGES.txt').read()
long_description = readme + '\n\n' + history

setup(name='Products.mcdutils',
      version='0.2b3',
      description='A Zope2 product which provides facilities for storing sessions in memcached.',
      long_description=long_description,
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Framework :: Zope2'],
      keywords='session memcache memcached Products',
      author='Tres Seaver',
      author_email='tseaver@palladion.com',
      maintainer='Christian Theune',
      maintainer_email='ct@gocept.com',
      url='https://bitbucket.org/ctheune/mcdutils',
      license='ZPL 2.1',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=['setuptools', 'Zope2'])
